@@ -35,9 +35,18 @@ import struct
 import time
 
 from decimal import Decimal
+from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
 
 from .constants import CharacterSet, FieldFlag, FieldType
 from .custom_types import HexLiteral
+from .types import (
+    DescriptionType,
+    RowType,
+    StrOrBytes,
+    ToMysqlInputTypes,
+    ToMysqlOutputTypes,
+    ToPythonOutputTypes,
+)
 from .utils import NUMERIC_TYPES
 
 CONVERT_ERROR = "Could not convert '{value}' to python {pytype}"
@@ -50,18 +59,25 @@ class MySQLConverterBase:
     be a subclass of this class.
     """
 
-    def __init__(self, charset="utf8", use_unicode=True, str_fallback=False):
-        self.python_types = None
-        self.mysql_types = None
-        self.charset = None
-        self.charset_id = 0
-        self.use_unicode = None
+    def __init__(
+        self,
+        charset: Optional[str] = "utf8",
+        use_unicode: bool = True,
+        str_fallback: bool = False,
+    ) -> None:
+        self.python_types: Optional[Tuple[Any, ...]] = None
+        self.mysql_types: Optional[Tuple[Any, ...]] = None
+        self.charset: Optional[str] = None
+        self.charset_id: int = 0
         self.set_charset(charset)
-        self.use_unicode = use_unicode
-        self.str_fallback = str_fallback
-        self._cache_field_types = {}
+        self.use_unicode: bool = use_unicode
+        self.str_fallback: bool = str_fallback
+        self._cache_field_types: Dict[
+            int,
+            Callable[[bytes, DescriptionType], ToPythonOutputTypes],
+        ] = {}
 
-    def set_charset(self, charset):
+    def set_charset(self, charset: Optional[str]) -> None:
         """Set character set"""
         if charset in ("utf8mb4", "utf8mb3"):
             charset = "utf8"
@@ -72,19 +88,26 @@ class MySQLConverterBase:
             self.charset = "utf8"
         self.charset_id = CharacterSet.get_charset_info(self.charset)[0]
 
-    def set_unicode(self, value=True):
+    def set_unicode(self, value: bool = True) -> None:
         """Set whether to use Unicode"""
         self.use_unicode = value
 
-    def to_mysql(self, value):
+    def to_mysql(
+        self, value: ToMysqlInputTypes
+    ) -> Union[ToMysqlInputTypes, HexLiteral]:
         """Convert Python data type to MySQL"""
         type_name = value.__class__.__name__.lower()
         try:
-            return getattr(self, f"_{type_name}_to_mysql")(value)
+            converted: ToMysqlOutputTypes = getattr(self, f"_{type_name}_to_mysql")(
+                value
+            )
+            return converted
         except AttributeError:
             return value
 
-    def to_python(self, vtype, value):
+    def to_python(
+        self, vtype: DescriptionType, value: Optional[bytes]
+    ) -> ToPythonOutputTypes:
         """Convert MySQL data type to Python"""
 
         if (value == b"\x00" or value is None) and vtype[1] != FieldType.BIT:
@@ -101,19 +124,20 @@ class MySQLConverterBase:
                 except AttributeError:
                     # We ignore field types which has no method
                     pass
-
+        if value is None:
+            return None
         try:
             return self._cache_field_types[vtype[1]](value, vtype)
         except KeyError:
             return value
 
     @staticmethod
-    def escape(value):
+    def escape(value: Any) -> Any:
         """Escape buffer for sending to MySQL"""
         return value
 
     @staticmethod
-    def quote(buf):
+    def quote(buf: Any) -> StrOrBytes:
         """Quote buffer for sending to MySQL"""
         return str(buf)
 
@@ -132,12 +156,20 @@ class MySQLConverter(MySQLConverterBase):
 
     """
 
-    def __init__(self, charset=None, use_unicode=True, str_fallback=False):
+    def __init__(
+        self,
+        charset: Optional[str] = None,
+        use_unicode: bool = True,
+        str_fallback: bool = False,
+    ) -> None:
         MySQLConverterBase.__init__(self, charset, use_unicode, str_fallback)
-        self._cache_field_types = {}
+        self._cache_field_types: Dict[
+            int,
+            Callable[[bytes, DescriptionType], ToPythonOutputTypes],
+        ] = {}
 
     @staticmethod
-    def escape(value):
+    def escape(value: Any) -> Any:
         """
         Escapes special characters as they are expected to by when MySQL
         receives them.
@@ -145,10 +177,6 @@ class MySQLConverter(MySQLConverterBase):
 
         Returns the value if not a string, or the escaped string.
         """
-        if value is None:
-            return value
-        if isinstance(value, NUMERIC_TYPES):
-            return value
         if isinstance(value, (bytes, bytearray)):
             value = value.replace(b"\\", b"\\\\")
             value = value.replace(b"\n", b"\\n")
@@ -156,7 +184,7 @@ class MySQLConverter(MySQLConverterBase):
             value = value.replace(b"\047", b"\134\047")  # single quotes
             value = value.replace(b"\042", b"\134\042")  # double quotes
             value = value.replace(b"\032", b"\134\032")  # for Win32
-        else:
+        elif isinstance(value, str) and not isinstance(value, HexLiteral):
             value = value.replace("\\", "\\\\")
             value = value.replace("\n", "\\n")
             value = value.replace("\r", "\\r")
@@ -166,7 +194,7 @@ class MySQLConverter(MySQLConverterBase):
         return value
 
     @staticmethod
-    def quote(buf):
+    def quote(buf: Optional[Union[float, int, Decimal, HexLiteral, bytes]]) -> bytes:
         """
         Quote the parameters for commands. General rules:
           o numbers are returns as bytes using ascii codec
@@ -179,13 +207,16 @@ class MySQLConverter(MySQLConverterBase):
             return str(buf).encode("ascii")
         if isinstance(buf, type(None)):
             return bytearray(b"NULL")
-        return bytearray(b"'" + buf + b"'")
+        return bytearray(b"'" + buf + b"'")  # type: ignore[operator]
 
-    def to_mysql(self, value):
+    def to_mysql(self, value: ToMysqlInputTypes) -> ToMysqlOutputTypes:
         """Convert Python data type to MySQL"""
         type_name = value.__class__.__name__.lower()
         try:
-            return getattr(self, f"_{type_name}_to_mysql")(value)
+            converted: ToMysqlOutputTypes = getattr(self, f"_{type_name}_to_mysql")(
+                value
+            )
+            return converted
         except AttributeError:
             if self.str_fallback:
                 return str(value).encode()
@@ -193,9 +224,14 @@ class MySQLConverter(MySQLConverterBase):
                 f"Python '{type_name}' cannot be converted to a MySQL type"
             ) from None
 
-    def to_python(self, vtype, value):
+    def to_python(
+        self,
+        vtype: DescriptionType,
+        value: Optional[bytes],
+    ) -> ToPythonOutputTypes:
         """Convert MySQL data type to Python"""
-        if value == 0 and vtype[1] != FieldType.BIT:  # \x00
+        # \x00
+        if value == 0 and vtype[1] != FieldType.BIT:
             # Don't go further when we hit a NULL value
             return None
         if value is None:
@@ -226,27 +262,31 @@ class MySQLConverter(MySQLConverterBase):
             raise TypeError(f"{err} (field {vtype[0]})") from err
 
     @staticmethod
-    def _int_to_mysql(value):
+    def _int_to_mysql(value: int) -> int:
         """Convert value to int"""
         return int(value)
 
     @staticmethod
-    def _long_to_mysql(value):
-        """Convert value to int"""
+    def _long_to_mysql(value: int) -> int:
+        """Convert value to int
+
+        Note: there is not type "long" in Python 3 since integers (int) are of unlimited size.
+        Since Python 2 is no longer supported, this method should be deprecated.
+        """
         return int(value)
 
     @staticmethod
-    def _float_to_mysql(value):
+    def _float_to_mysql(value: float) -> Optional[float]:
         """Convert value to float"""
         if math.isnan(value):
             return None
         return float(value)
 
-    def _str_to_mysql(self, value):
+    def _str_to_mysql(self, value: str) -> Union[bytes, HexLiteral]:
         """Convert value to string"""
         return self._unicode_to_mysql(value)
 
-    def _unicode_to_mysql(self, value):
+    def _unicode_to_mysql(self, value: str) -> Union[bytes, HexLiteral]:
         """Convert unicode"""
         charset = self.charset
         charset_id = self.charset_id
@@ -260,24 +300,22 @@ class MySQLConverter(MySQLConverterBase):
         return encoded
 
     @staticmethod
-    def _bytes_to_mysql(value):
+    def _bytes_to_mysql(value: bytes) -> bytes:
         """Convert value to bytes"""
         return value
 
     @staticmethod
-    def _bytearray_to_mysql(value):
+    def _bytearray_to_mysql(value: bytearray) -> bytes:
         """Convert value to bytes"""
         return bytes(value)
 
     @staticmethod
-    def _bool_to_mysql(value):
+    def _bool_to_mysql(value: bool) -> int:
         """Convert value to boolean"""
-        if value:
-            return 1
-        return 0
+        return 1 if value else 0
 
     @staticmethod
-    def _nonetype_to_mysql(value):  # pylint: disable=unused-argument
+    def _nonetype_to_mysql(value: None) -> None:  # pylint: disable=unused-argument
         """
         This would return what None would be in MySQL, but instead we
         leave it None and return it right away. The actual conversion
@@ -288,7 +326,7 @@ class MySQLConverter(MySQLConverterBase):
         return None
 
     @staticmethod
-    def _datetime_to_mysql(value):
+    def _datetime_to_mysql(value: datetime.datetime) -> bytes:
         """
         Converts a datetime instance to a string suitable for MySQL.
         The returned string has format: %Y-%m-%d %H:%M:%S[.%f]
@@ -320,7 +358,7 @@ class MySQLConverter(MySQLConverterBase):
         ).encode("ascii")
 
     @staticmethod
-    def _date_to_mysql(value):
+    def _date_to_mysql(value: datetime.date) -> bytes:
         """
         Converts a date instance to a string suitable for MySQL.
         The returned string has format: %Y-%m-%d
@@ -332,7 +370,7 @@ class MySQLConverter(MySQLConverterBase):
         return f"{value.year:04d}-{value.month:02d}-{value.day:02d}".encode("ascii")
 
     @staticmethod
-    def _time_to_mysql(value):
+    def _time_to_mysql(value: datetime.time) -> bytes:
         """
         Converts a time instance to a string suitable for MySQL.
         The returned string has format: %H:%M:%S[.%f]
@@ -346,7 +384,7 @@ class MySQLConverter(MySQLConverterBase):
         return value.strftime("%H:%M:%S").encode("ascii")
 
     @staticmethod
-    def _struct_time_to_mysql(value):
+    def _struct_time_to_mysql(value: time.struct_time) -> bytes:
         """
         Converts a time.struct_time sequence to a string suitable
         for MySQL.
@@ -357,7 +395,7 @@ class MySQLConverter(MySQLConverterBase):
         return time.strftime("%Y-%m-%d %H:%M:%S", value).encode("ascii")
 
     @staticmethod
-    def _timedelta_to_mysql(value):
+    def _timedelta_to_mysql(value: datetime.timedelta) -> bytes:
         """
         Converts a timedelta instance to a string suitable for MySQL.
         The returned string has format: %H:%M:%S
@@ -390,7 +428,7 @@ class MySQLConverter(MySQLConverterBase):
         return result.encode("ascii")
 
     @staticmethod
-    def _decimal_to_mysql(value):
+    def _decimal_to_mysql(value: Decimal) -> Optional[bytes]:
         """
         Converts a decimal.Decimal instance to a string suitable for
         MySQL.
@@ -402,7 +440,9 @@ class MySQLConverter(MySQLConverterBase):
 
         return None
 
-    def row_to_python(self, row, fields):
+    def row_to_python(
+        self, row: Tuple[bytes, ...], fields: List[DescriptionType]
+    ) -> RowType:
         """Convert a MySQL text result row to Python types
 
         The row argument is a sequence containing text result returned
@@ -412,7 +452,7 @@ class MySQLConverter(MySQLConverterBase):
         Returns a tuple.
         """
         i = 0
-        result = [None] * len(fields)
+        result: List[ToPythonOutputTypes] = [None] * len(fields)
 
         if not self._cache_field_types:
             self._cache_field_types = {}
@@ -442,7 +482,8 @@ class MySQLConverter(MySQLConverterBase):
                 except UnicodeDecodeError:
                     result[i] = row[i]
             except (ValueError, TypeError) as err:
-                err.message = f"{err} (field {field[0]})"
+                # Item "ValueError" of "Union[ValueError, TypeError]" has no attribute "message"
+                err.message = f"{err} (field {field[0]})"  # type: ignore[union-attr]
                 raise
 
             i += 1
@@ -451,7 +492,7 @@ class MySQLConverter(MySQLConverterBase):
 
     # pylint: disable=unused-argument
     @staticmethod
-    def _float_to_python(value, desc=None):
+    def _float_to_python(value: bytes, desc: Optional[DescriptionType] = None) -> float:
         """
         Returns value as float type.
         """
@@ -460,7 +501,7 @@ class MySQLConverter(MySQLConverterBase):
     _double_to_python = _float_to_python
 
     @staticmethod
-    def _int_to_python(value, desc=None):
+    def _int_to_python(value: bytes, desc: Optional[DescriptionType] = None) -> int:
         """
         Returns value as int type.
         """
@@ -472,7 +513,9 @@ class MySQLConverter(MySQLConverterBase):
     _long_to_python = _int_to_python
     _longlong_to_python = _int_to_python
 
-    def _decimal_to_python(self, value, desc=None):
+    def _decimal_to_python(
+        self, value: bytes, desc: Optional[DescriptionType] = None
+    ) -> Decimal:
         """
         Returns value as a decimal.Decimal.
         """
@@ -482,22 +525,24 @@ class MySQLConverter(MySQLConverterBase):
     _newdecimal_to_python = _decimal_to_python
 
     @staticmethod
-    def _str(value, desc=None):
+    def _str(value: bytes, desc: Optional[DescriptionType] = None) -> str:
         """
         Returns value as str type.
         """
         return str(value)
 
     @staticmethod
-    def _bit_to_python(value, dsc=None):
+    def _bit_to_python(value: bytes, dsc: Optional[DescriptionType] = None) -> int:
         """Returns BIT columntype as integer"""
         int_val = value
         if len(int_val) < 8:
             int_val = b"\x00" * (8 - len(int_val)) + int_val
-        return struct.unpack(">Q", int_val)[0]
+        return int(struct.unpack(">Q", int_val)[0])
 
     @staticmethod
-    def _date_to_python(value, dsc=None):
+    def _date_to_python(
+        value: bytes, dsc: Optional[DescriptionType] = None
+    ) -> Optional[datetime.date]:
         """Converts TIME column MySQL to a python datetime.datetime type.
 
         Raises ValueError if the value can not be converted.
@@ -516,13 +561,15 @@ class MySQLConverter(MySQLConverterBase):
                 return None
         except (IndexError, ValueError):
             raise ValueError(
-                f"Could not convert {value} to python datetime.timedelta"
+                f"Could not convert {repr(value)} to python datetime.timedelta"
             ) from None
 
     _NEWDATE_to_python = _date_to_python
 
     @staticmethod
-    def _time_to_python(value, dsc=None):
+    def _time_to_python(
+        value: bytes, dsc: Optional[DescriptionType] = None
+    ) -> datetime.timedelta:
         """Converts TIME column value to python datetime.time value type.
 
         Converts the TIME column MySQL type passed as bytes to a python
@@ -530,8 +577,9 @@ class MySQLConverter(MySQLConverterBase):
 
         Raises ValueError if the value can not be converted.
 
-        Returns datetime.time type.
+        Returns datetime.timedelta type.
         """
+        mcs: Optional[Union[int, bytes]] = None
         try:
             (hms, mcs) = value.split(b".")
             mcs = int(mcs.ljust(6, b"0"))
@@ -541,7 +589,11 @@ class MySQLConverter(MySQLConverterBase):
         try:
             (hours, mins, secs) = [int(d) for d in hms.split(b":")]
             if value[0] == 45 or value[0] == "-":
-                mins, secs, mcs = -mins, -secs, -mcs
+                mins, secs, mcs = (
+                    -mins,
+                    -secs,
+                    -mcs,  # pylint: disable=invalid-unary-operand-type
+                )
             return datetime.timedelta(
                 hours=hours, minutes=mins, seconds=secs, microseconds=mcs
             )
@@ -551,7 +603,9 @@ class MySQLConverter(MySQLConverterBase):
             ) from None
 
     @staticmethod
-    def _datetime_to_python(value, dsc=None):
+    def _datetime_to_python(
+        value: bytes, dsc: Optional[DescriptionType] = None
+    ) -> Optional[datetime.datetime]:
         """Converts DATETIME column value to python datetime.time value type.
 
         Converts the DATETIME column MySQL type passed as bytes to a python
@@ -562,6 +616,7 @@ class MySQLConverter(MySQLConverterBase):
         if isinstance(value, datetime.datetime):
             return value
         datetime_val = None
+        mcs: Optional[Union[int, bytes]] = None
         try:
             (date_, time_) = value.split(b" ")
             if len(time_) > 8:
@@ -584,7 +639,7 @@ class MySQLConverter(MySQLConverterBase):
             # Traditionaly C/py returns None for this well formed but
             # invalid datetime for python like '0000-00-00 HH:MM:SS'.
             try:
-                datetime_val = datetime.datetime(*dtval)
+                datetime_val = datetime.datetime(*dtval)  # type: ignore[arg-type]
             except ValueError:
                 return None
         except (IndexError, TypeError):
@@ -597,16 +652,18 @@ class MySQLConverter(MySQLConverterBase):
     _timestamp_to_python = _datetime_to_python
 
     @staticmethod
-    def _year_to_python(value, desc=None):
+    def _year_to_python(value: bytes, dsc: Optional[DescriptionType] = None) -> int:
         """Returns YEAR column type as integer"""
         try:
             year = int(value)
         except ValueError as err:
-            raise ValueError(f"Failed converting YEAR to int ({value})") from err
+            raise ValueError(f"Failed converting YEAR to int ({repr(value)})") from err
 
         return year
 
-    def _set_to_python(self, value, dsc=None):
+    def _set_to_python(
+        self, value: bytes, dsc: Optional[DescriptionType] = None
+    ) -> Set[str]:
         """Returns SET column type as set
 
         Actually, MySQL protocol sees a SET as a string type field. So this
@@ -621,10 +678,14 @@ class MySQLConverter(MySQLConverterBase):
         try:
             set_type = set(val.split(","))
         except ValueError as err:
-            raise ValueError(f"Could not convert set {value} to a sequence") from err
+            raise ValueError(
+                f"Could not convert set {repr(value)} to a sequence"
+            ) from err
         return set_type
 
-    def _string_to_python(self, value, dsc=None):
+    def _string_to_python(
+        self, value: bytes, dsc: Optional[DescriptionType] = None
+    ) -> Union[StrOrBytes, Set[str]]:
         """
         Note that a SET is a string too, but using the FieldFlag we can see
         whether we have to split it.
@@ -638,7 +699,8 @@ class MySQLConverter(MySQLConverterBase):
                 return value.decode(self.charset)
             if dsc[7] & FieldFlag.SET:
                 return self._set_to_python(value, dsc)
-            if dsc[8] == 63:  # 'binary' charset
+            # 'binary' charset
+            if dsc[8] == 63:
                 return value
         if isinstance(value, (bytes, bytearray)) and self.use_unicode:
             try:
@@ -651,13 +713,16 @@ class MySQLConverter(MySQLConverterBase):
     _var_string_to_python = _string_to_python
     _json_to_python = _string_to_python
 
-    def _blob_to_python(self, value, dsc=None):
+    def _blob_to_python(
+        self, value: bytes, dsc: Optional[DescriptionType] = None
+    ) -> Union[StrOrBytes, Set[str]]:
         """Convert BLOB data type to Python."""
         if dsc is not None:
             if (
                 dsc[7] & FieldFlag.BLOB
                 and dsc[7] & FieldFlag.BINARY
-                and dsc[8] == 63  # 'binary' charset
+                # 'binary' charset
+                and dsc[8] == 63
             ):
                 return bytes(value)
         return self._string_to_python(value, dsc)
