@@ -1,4 +1,4 @@
-# Copyright (c) 2016, 2020, Oracle and/or its affiliates. All rights reserved.
+# Copyright (c) 2016, 2022, Oracle and/or its affiliates.
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License, version 2.0, as
@@ -32,56 +32,68 @@ import json
 import warnings
 
 from .dbdoc import DbDoc
-from .compat import STRING_TYPES
-from .errorcode import (ER_NO_SUCH_TABLE, ER_TABLE_EXISTS_ERROR,
-                        ER_X_CMD_NUM_ARGUMENTS, ER_X_INVALID_ADMIN_COMMAND)
+from .errorcode import (
+    ER_NO_SUCH_TABLE,
+    ER_TABLE_EXISTS_ERROR,
+    ER_X_CMD_NUM_ARGUMENTS,
+    ER_X_INVALID_ADMIN_COMMAND,
+)
 from .errors import NotSupportedError, OperationalError, ProgrammingError
 from .helpers import deprecated, escape, quote_identifier
-from .statement import (FindStatement, AddStatement, RemoveStatement,
-                        ModifyStatement, SelectStatement, InsertStatement,
-                        DeleteStatement, UpdateStatement,
-                        CreateCollectionIndexStatement)
+from .statement import (
+    AddStatement,
+    CreateCollectionIndexStatement,
+    DeleteStatement,
+    FindStatement,
+    InsertStatement,
+    ModifyStatement,
+    RemoveStatement,
+    SelectStatement,
+    UpdateStatement,
+)
 
-
-_COUNT_VIEWS_QUERY = ("SELECT COUNT(*) FROM information_schema.views "
-                      "WHERE table_schema = '{0}' AND table_name = '{1}'")
-_COUNT_TABLES_QUERY = ("SELECT COUNT(*) FROM information_schema.tables "
-                       "WHERE table_schema = '{0}' AND table_name = '{1}'")
-_COUNT_SCHEMAS_QUERY = ("SELECT COUNT(*) FROM information_schema.schemata "
-                        "WHERE schema_name = '{0}'")
+_COUNT_VIEWS_QUERY = (
+    "SELECT COUNT(*) FROM information_schema.views "
+    "WHERE table_schema = '{0}' AND table_name = '{1}'"
+)
+_COUNT_TABLES_QUERY = (
+    "SELECT COUNT(*) FROM information_schema.tables "
+    "WHERE table_schema = '{0}' AND table_name = '{1}'"
+)
+_COUNT_SCHEMAS_QUERY = (
+    "SELECT COUNT(*) FROM information_schema.schemata WHERE schema_name = '{0}'"
+)
 _COUNT_QUERY = "SELECT COUNT(*) FROM {0}.{1}"
 _DROP_TABLE_QUERY = "DROP TABLE IF EXISTS {0}.{1}"
 
 
-class DatabaseObject(object):
+class DatabaseObject:
     """Provides base functionality for database objects.
 
     Args:
         schema (mysqlx.Schema): The Schema object.
         name (str): The database object name.
     """
+
     def __init__(self, schema, name):
         self._schema = schema
-        self._name = name
+        self._name = name.decode() if isinstance(name, bytes) else name
         self._session = self._schema.get_session()
         self._connection = self._session.get_connection()
 
     @property
     def session(self):
-        """:class:`mysqlx.Session`: The Session object.
-        """
+        """:class:`mysqlx.Session`: The Session object."""
         return self._session
 
     @property
     def schema(self):
-        """:class:`mysqlx.Schema`: The Schema object.
-        """
+        """:class:`mysqlx.Schema`: The Schema object."""
         return self._schema
 
     @property
     def name(self):
-        """str: The name of this database object.
-        """
+        """str: The name of this database object."""
         return self._name
 
     def get_connection(self):
@@ -163,9 +175,10 @@ class Schema(DatabaseObject):
         session (mysqlx.XSession): Session object.
         name (str): The Schema name.
     """
+
     def __init__(self, session, name):
         self._session = session
-        super(Schema, self).__init__(self, name)
+        super().__init__(self, name)
 
     def exists_in_database(self):
         """Verifies if this object exists in the database.
@@ -182,8 +195,7 @@ class Schema(DatabaseObject):
         Returns:
             `list`: List of Collection objects.
         """
-        rows = self._connection.get_row_result("list_objects",
-                                               {"schema": self._name})
+        rows = self._connection.get_row_result("list_objects", {"schema": self._name})
         rows.fetch_all()
         collections = []
         for row in rows:
@@ -211,11 +223,13 @@ class Schema(DatabaseObject):
         Returns:
             `list`: List of Table objects.
         """
-        rows = self._connection.get_row_result("list_objects",
-                                               {"schema": self._name})
+        rows = self._connection.get_row_result("list_objects", {"schema": self._name})
         rows.fetch_all()
         tables = []
-        object_types = ("TABLE", "VIEW",)
+        object_types = (
+            "TABLE",
+            "VIEW",
+        )
         for row in rows:
             if row["type"] in object_types:
                 try:
@@ -268,11 +282,14 @@ class Schema(DatabaseObject):
             name (str): The name of the collection to be dropped.
         """
         self._connection.execute_nonquery(
-            "sql", _DROP_TABLE_QUERY.format(quote_identifier(self._name),
-                                            quote_identifier(name)), False)
+            "sql",
+            _DROP_TABLE_QUERY.format(
+                quote_identifier(self._name), quote_identifier(name)
+            ),
+            False,
+        )
 
-    def create_collection(self, name, reuse_existing=False, validation=None,
-                          **kwargs):
+    def create_collection(self, name, reuse_existing=False, validation=None, **kwargs):
         """Creates in the current schema a new collection with the specified
         name and retrieves an object representing the new collection created.
 
@@ -300,9 +317,11 @@ class Schema(DatabaseObject):
             raise ProgrammingError("Collection name is invalid")
 
         if "reuse" in kwargs:
-            warnings.warn("'reuse' is deprecated since 8.0.21. "
-                          "Please use 'reuse_existing' instead",
-                          DeprecationWarning)
+            warnings.warn(
+                "'reuse' is deprecated since 8.0.21. "
+                "Please use 'reuse_existing' instead",
+                DeprecationWarning,
+            )
             reuse_existing = kwargs["reuse"]
 
         collection = Collection(self, name)
@@ -315,8 +334,7 @@ class Schema(DatabaseObject):
             valid_options = ("level", "schema")
             for option in validation:
                 if option not in valid_options:
-                    raise ProgrammingError("Invalid option in 'validation': {}"
-                                           "".format(option))
+                    raise ProgrammingError(f"Invalid option in 'validation': {option}")
 
             options = []
 
@@ -331,26 +349,32 @@ class Schema(DatabaseObject):
                 if not isinstance(schema, (str, dict)):
                     raise ProgrammingError("Invalid value for 'schema'")
                 options.append(
-                    ("schema", json.dumps(schema)
-                               if isinstance(schema, dict) else schema))
+                    (
+                        "schema",
+                        json.dumps(schema) if isinstance(schema, dict) else schema,
+                    )
+                )
 
             fields["options"] = ("validation", options)
 
         try:
             self._connection.execute_nonquery(
-                "mysqlx", "create_collection", True, fields)
+                "mysqlx", "create_collection", True, fields
+            )
         except OperationalError as err:
             if err.errno == ER_X_CMD_NUM_ARGUMENTS:
                 raise NotSupportedError(
                     "Your MySQL server does not support the requested "
                     "operation. Please update to MySQL 8.0.19 or a later "
-                    "version")
+                    "version"
+                ) from err
             if err.errno == ER_TABLE_EXISTS_ERROR:
                 if not reuse_existing:
                     raise ProgrammingError(
-                        "Collection '{}' already exists".format(name))
+                        f"Collection '{name}' already exists"
+                    ) from err
             else:
-                raise ProgrammingError(err.msg, err.errno)
+                raise ProgrammingError(err.msg, err.errno) from err
 
         return collection
 
@@ -381,8 +405,7 @@ class Schema(DatabaseObject):
         valid_options = ("level", "schema")
         for option in validation:
             if option not in valid_options:
-                raise ProgrammingError("Invalid option in 'validation': {}"
-                                       "".format(option))
+                raise ProgrammingError(f"Invalid option in 'validation': {option}")
         options = []
 
         if "level" in validation:
@@ -396,25 +419,30 @@ class Schema(DatabaseObject):
             if not isinstance(schema, (str, dict)):
                 raise ProgrammingError("Invalid value for 'schema'")
             options.append(
-                ("schema", json.dumps(schema)
-                           if isinstance(schema, dict) else schema))
+                (
+                    "schema",
+                    json.dumps(schema) if isinstance(schema, dict) else schema,
+                )
+            )
 
         fields = {
             "schema": self._name,
             "name": name,
-            "options": ("validation", options)
+            "options": ("validation", options),
         }
 
         try:
             self._connection.execute_nonquery(
-                "mysqlx", "modify_collection_options", True, fields)
+                "mysqlx", "modify_collection_options", True, fields
+            )
         except OperationalError as err:
             if err.errno == ER_X_INVALID_ADMIN_COMMAND:
                 raise NotSupportedError(
                     "Your MySQL server does not support the requested "
                     "operation. Please update to MySQL 8.0.19 or a later "
-                    "version")
-            raise ProgrammingError(err.msg, err.errno)
+                    "version"
+                ) from err
+            raise ProgrammingError(err.msg, err.errno) from err
 
 
 class Collection(DatabaseObject):
@@ -431,8 +459,7 @@ class Collection(DatabaseObject):
         Returns:
             bool: `True` if object exists in database.
         """
-        sql = _COUNT_TABLES_QUERY.format(escape(self._schema.name),
-                                         escape(self._name))
+        sql = _COUNT_TABLES_QUERY.format(escape(self._schema.name), escape(self._name))
         return self._connection.execute_sql_scalar(sql) == 1
 
     def find(self, condition=None):
@@ -497,15 +524,17 @@ class Collection(DatabaseObject):
         Returns:
             int: The total of documents in the collection.
         """
-        sql = _COUNT_QUERY.format(quote_identifier(self._schema.name),
-                                  quote_identifier(self._name))
+        sql = _COUNT_QUERY.format(
+            quote_identifier(self._schema.name), quote_identifier(self._name)
+        )
         try:
             res = self._connection.execute_sql_scalar(sql)
         except OperationalError as err:
             if err.errno == ER_NO_SUCH_TABLE:
                 raise OperationalError(
-                    "Collection '{}' does not exist in schema '{}'"
-                    "".format(self._name, self._schema.name))
+                    f"Collection '{self._name}' does not exist in schema "
+                    f"'{self._schema.name}'"
+                ) from err
             raise
         return res
 
@@ -539,10 +568,16 @@ class Collection(DatabaseObject):
         Args:
             index_name (str): Index name.
         """
-        self._connection.execute_nonquery("mysqlx", "drop_collection_index",
-                                          False, {"schema": self._schema.name,
-                                                  "collection": self._name,
-                                                  "name": index_name})
+        self._connection.execute_nonquery(
+            "mysqlx",
+            "drop_collection_index",
+            False,
+            {
+                "schema": self._schema.name,
+                "collection": self._name,
+                "name": index_name,
+            },
+        )
 
     def replace_one(self, doc_id, doc):
         """Replaces the Document matching the document ID with a new document
@@ -552,8 +587,12 @@ class Collection(DatabaseObject):
             doc_id (str): Document ID
             doc (:class:`mysqlx.DbDoc` or `dict`): New Document
         """
-        return self.modify("_id = :id").set("$", doc) \
-                   .bind("id", doc_id).execute()
+        if "_id" in doc and doc["_id"] != doc_id:
+            raise ProgrammingError(
+                "Replacement document has an _id that is different than the "
+                "matched document"
+            )
+        return self.modify("_id = :id").set("$", doc).bind("id", doc_id).execute()
 
     def add_or_replace_one(self, doc_id, doc):
         """Upserts the Document matching the document ID with a new document
@@ -563,6 +602,11 @@ class Collection(DatabaseObject):
             doc_id (str): Document ID
             doc (:class:`mysqlx.DbDoc` or dict): New Document
         """
+        if "_id" in doc and doc["_id"] != doc_id:
+            raise ProgrammingError(
+                "Replacement document has an _id that is different than the "
+                "matched document"
+            )
         if not isinstance(doc, DbDoc):
             doc = DbDoc(doc)
         return self.add(doc.copy(doc_id)).upsert(True).execute()
@@ -610,8 +654,7 @@ class Table(DatabaseObject):
         Returns:
             bool: `True` if object exists in database.
         """
-        sql = _COUNT_TABLES_QUERY.format(escape(self._schema.name),
-                                         escape(self._name))
+        sql = _COUNT_TABLES_QUERY.format(escape(self._schema.name), escape(self._name))
         return self._connection.execute_sql_scalar(sql) == 1
 
     def select(self, *fields):
@@ -669,15 +712,17 @@ class Table(DatabaseObject):
         Returns:
             int: The total of rows in the table.
         """
-        sql = _COUNT_QUERY.format(quote_identifier(self._schema.name),
-                                  quote_identifier(self._name))
+        sql = _COUNT_QUERY.format(
+            quote_identifier(self._schema.name), quote_identifier(self._name)
+        )
         try:
             res = self._connection.execute_sql_scalar(sql)
         except OperationalError as err:
             if err.errno == ER_NO_SUCH_TABLE:
                 raise OperationalError(
-                    "Table '{}' does not exist in schema '{}'"
-                    "".format(self._name, self._schema.name))
+                    f"Table '{self._name}' does not exist in schema "
+                    f"'{self._schema.name}'"
+                ) from err
             raise
         return res
 
@@ -687,8 +732,7 @@ class Table(DatabaseObject):
         Returns:
             bool: `True` if the underlying object is a view.
         """
-        sql = _COUNT_VIEWS_QUERY.format(escape(self._schema.name),
-                                        escape(self._name))
+        sql = _COUNT_VIEWS_QUERY.format(escape(self._schema.name), escape(self._name))
         return self._connection.execute_sql_scalar(sql) == 1
 
 
@@ -708,6 +752,5 @@ class View(Table):
         Returns:
             bool: `True` if object exists in database.
         """
-        sql = _COUNT_VIEWS_QUERY.format(escape(self._schema.name),
-                                        escape(self._name))
+        sql = _COUNT_VIEWS_QUERY.format(escape(self._schema.name), escape(self._name))
         return self._connection.execute_sql_scalar(sql) == 1
